@@ -21,6 +21,11 @@ ods['gyrokinetics'].load(sample_filename)
 pprint(ods.pretty_paths())
 
 # save a copy
+try:
+    __file__
+except NameError:
+    import inspect
+    __file__  = inspect.getfile(lambda: None)
 filename = omas_testdir(__file__) + '/gkdb_linear_initialvalue.json'
 ods['gyrokinetics'].save(filename)
 
